#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef struct pessoa_t {
	struct pessoa_vtable_t *vtable;
	char nome[16];
	int idade;
	char sexo;
} pessoa_t;

// Implementação do método 'imprime()' de um objeto do tipo 'pessoa_t'.
void imprime_pessoa(pessoa_t *p) {
	printf("Esta pessoa se chama %s, e tem %d anos.\n", p->nome, p->idade);
}

// Implementação do método 'cumprimenta()' de um objeto do tipo 'pessoa_t'.
void cumprimenta_pessoa(pessoa_t *p) {
	printf("Olá, %s!\n", p->nome);
}

// Tipo da vtable de pessoa_t.
typedef struct pessoa_vtable_t {
	void (*imprime)(pessoa_t *);
	void (*cumprimenta)(pessoa_t *);
} pessoa_vtable_t;

// A vtable propriamente dita.
pessoa_vtable_t pessoa_vtable = { imprime_pessoa, cumprimenta_pessoa };

pessoa_t *make_pessoa(char *nome, int idade, char sexo) {
	pessoa_t *new = malloc(sizeof(pessoa_t));
	strcpy(new->nome, nome);
	new->idade = idade;
	new->sexo = sexo;

	new->vtable = &pessoa_vtable;
	return new;
}

void imprime(pessoa_t *p)     { p->vtable->imprime(p); }
void cumprimenta(pessoa_t *p) { p->vtable->cumprimenta(p); }

typedef struct funcionario_t {
	struct funcionario_vtable_t *vtable;
	char nome[16];
	int idade;
	char sexo;
	float salario;
} funcionario_t;

typedef struct funcionario_vtable_t {
	void (*imprime)(funcionario_t *);
	void (*cumprimenta)(funcionario_t *);
	void (*multiplica_salario)(funcionario_t *, float fator);
} funcionario_vtable_t;

void imprime_funcionario(funcionario_t *f) {
	imprime_pessoa(f);   // Chamamos o método da superclasse.
	printf("Esta criatura ganha %f dinheiros.\n", f->salario);
}

void multiplica_salario_funcionario(funcionario_t *f, float fator) {
	f->salario *= fator;
}

funcionario_vtable_t funcionario_vtable =
    { imprime_funcionario, cumprimenta_pessoa, multiplica_salario_funcionario };

funcionario_t *make_funcionario(char *nome, int idade, char sexo, float salario) {
	funcionario_t *new = malloc(sizeof(funcionario_t));
	strcpy(new->nome, nome);
	new->idade = idade;
	new->sexo = sexo;
	new->salario = salario;

	new->vtable = &funcionario_vtable;
	return new;
}

void multiplica_salario(funcionario_t *f, float fator) {
	f->vtable->multiplica_salario(f, fator);
}

int main() {
	pessoa_t *p1 = make_pessoa("Hildur", 18, 'f');
	funcionario_t *p2 = make_funcionario("Elmord", 20, 'm', 1000);

	imprime(p1);
	imprime(p2);
	cumprimenta(p1);
	cumprimenta(p2);
	multiplica_salario(p2, 1.5);
	imprime(p2);

	return 0;
} 
